import abc
import inspect
import functools
from typing import List, get_origin, get_args
from types import GenericAlias

__COMPONENT_CLASS = []
__NAME_COMPONENT = {}
__COMPONENT_INSTANCE = {}


class InjectableError(Exception):
    def __init__(self, msg):
        self.__msg = msg

    def __str__(self):
        return self.__msg


def component(cls=None, bean_name=None):
    if cls is None:
        return functools.partial(component, bean_name=bean_name)

    __COMPONENT_CLASS.append(cls)
    if bean_name:
        if bean_name in __NAME_COMPONENT:
            raise InjectableError(f"bean name[{bean_name}] is already defined.")
        __NAME_COMPONENT[bean_name] = cls
    return cls


def autowired(func=None):
    if func is None:
        return functools.partial(autowired)

    @functools.wraps(func)
    def wrap(*args, **kwargs):
        self_type = type(args[0])
        if self_type in __COMPONENT_CLASS and self_type not in __COMPONENT_INSTANCE:
            __COMPONENT_INSTANCE[self_type] = args[0]
        for p_name, p_type in inspect.signature(func).parameters.items():
            if p_name == 'self' or p_type == inspect.Parameter.empty:
                continue
            if get_origin(p_type.annotation) is list:
                inject_types = get_args(p_type.annotation)
                if len(inject_types) > 0:
                    instances = [get_instance(x) for x in __COMPONENT_CLASS if issubclass(x, inject_types)]
                    if len(instances) > 0:
                        kwargs[p_name] = instances
            else:
                components = [x for x in __COMPONENT_CLASS if issubclass(x, p_type.annotation)]
                if len(components) > 0:
                    kwargs[p_name] = get_instance(components[0])
        func(*args, **kwargs)

    return wrap


def get_instance(t):
    if t not in __COMPONENT_CLASS:
        return None
    if t not in __COMPONENT_INSTANCE:
        __COMPONENT_INSTANCE[t] = t()
    return __COMPONENT_INSTANCE[t]


class BaseBean(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def do_something(self):
        pass


@component(bean_name="one")
class OneImplBean(BaseBean):

    def do_something(self):
        print("one bean")


@component(bean_name="two")
class TwoImplBean(BaseBean):

    def do_something(self):
        print("two bean")


class ServiceBean:
    @autowired
    def __init__(self, base: BaseBean = None, bases: List[BaseBean] = None):
        base.do_something()
        for b in bases:
            b.do_something()




print(__COMPONENT_CLASS, __NAME_COMPONENT)

if __name__ == '__main__':
    ServiceBean()
